from pymongo import MongoClient, errors as MongoErrors
from pymongo.database import Database as MongoDatabase
from jasmin_telnet.proxy import Proxy as JasminTelnetProxy

import logging


class MongoDB:

    def __init__(self, connection_string, database_name):
        """ Constructor """
        self.connection_string = connection_string
        self.database_name = database_name
        logging.info("Starting MongoDB cluster's connection")

    def logger_callback(self, msg: str):
        logging.info(msg=msg)

    def startConnection(self) -> bool:
        mongoclient = MongoClient(self.connection_string)
        server_info = mongoclient.server_info()
        if isinstance(server_info, dict) and 'ok' in server_info and server_info['ok'] == 1:
            logging.info("Connected to MongoDB")
            self.mongoclient = mongoclient
            database: MongoDatabase = self.mongoclient[self.database_name]
            database_info = database.command("buildinfo")
            if isinstance(database_info, dict) and 'ok' in database_info and database_info['ok'] == 1:
                logging.info(f"Set to use database: {self.database_name}")
                logging.info("")
                self.database = database
                return True
            else:
                logging.critical(
                    f"Failed to use database: {self.database_name}")
                return False
        else:
            logging.critical("Failed to connect to MongoDB")
            return False

    def get_one_module(self, module: str) -> dict[str, str | float | bool]:
        "" ""
        data: dict[str, str | float | bool] = {}
        cursor = self.database[module].find()
        for row in cursor:
            sub_id = row["_id"]
            del row["_id"]
            data[sub_id] = row

        return data

    def get_one_submodule(self, module: str, sub_id: str) -> dict[str, str | float | bool]:
        "" ""
        return self.database[module].find_one({"_id": sub_id})

    def insert_one(self, module, sub_id, data):
        data["_id"] = sub_id
        self.database[module].insert_one(data)

    def increment_one(self, module, sub_id, data):
        self.database[module].update_one(
            {"_id": sub_id}, {'$inc': data})

    def update_one(self, module, sub_id, data, upsert=True):
        self.database[module].update_one(
            {"_id": sub_id}, {'$set': data}, upsert=upsert)

    def pullAllConfigurations(self, telnet_config: dict):
        "" ""
        sync_data = {
            "smppccm": {},
            "httpccm": {},
            "group": {},
            "user": {},
            "filter": {},
            "mointerceptor": {},
            "mtinterceptor": {},
            "morouter": {},
            "mtrouter": {},
        }

        for module in sync_data.keys():
            cursor = self.database[module].find()
            for row in cursor:
                sub_id = row["_id"]
                del row["_id"]
                sync_data[module][sub_id] = row

        self.stream_handler(
            telnet_config=telnet_config,
            payload={
                "change_type": "sync",
                "sync_data": sync_data
            })

    def stream(self, telnet_config: dict, syncCurrentFirst: bool = False):
        self.syncCurrentFirst = syncCurrentFirst
        try:
            with self.database.watch(full_document='updateLookup') as stream:

                if self.syncCurrentFirst is True:
                    # Sync current data to Jasmin before waiting for changes
                    logging.info("Sync current configuration first is ENABLED")
                    self.pullAllConfigurations(telnet_config=telnet_config)
                    logging.info("")
                else:
                    logging.info("Sync current configuration first is DISABLED")
                    logging.info("Skipping synchronizing current configurations")
                    logging.info("")


                logging.info("Starting MongoDB cluster's Change Stream")
                logging.info("")
                for change in stream:
                    module = change["ns"]["coll"]
                    sub_id = change["documentKey"]["_id"]
                    change_type = change["operationType"]
                    if change_type == "invalidate":
                        return
                    if change_type in ["create", "drop", "dropDatabase", "rename", "modify"]:
                        self.pullAllConfigurations(telnet_config=telnet_config)
                        continue
                    if "updateDescription" in change and module in ["user", "smppccm"]:
                        sub_data = change["updateDescription"]["updatedFields"]
                    elif "fullDocument" in change:
                        sub_data = change["fullDocument"]
                    else:
                        sub_data = {}

                    if isinstance(sub_data, dict) and "_id" in sub_data:
                        del sub_data["_id"]

                    self.stream_handler(telnet_config=telnet_config,
                                        payload={
                                            "module": module,
                                            "sub_id": sub_id,
                                            "change_type": change_type,
                                            "sub_data": sub_data
                                        })
        except MongoErrors.PyMongoError as err:
            # The ChangeStream encountered an unrecoverable error or the
            # resume attempt failed to recreate the cursor.
            logging.error(err)
            pass
        except KeyboardInterrupt:
            pass

    def stream_handler(self, telnet_config: dict, payload):
        """Handle callback for jasmin stream"""
        jasmin_proxy = JasminTelnetProxy(
            host=telnet_config["host"],
            port=telnet_config["port"],
            timeout=telnet_config["timeout"],
            auth=telnet_config["auth"],
            username=telnet_config["username"],
            password=telnet_config["password"],
            standard_prompt=telnet_config["standard_prompt"],
            interactive_prompt=telnet_config["interactive_prompt"],
            log_status=True,
            logger=self.logger_callback
        )
        if hasattr(self, "syncCurrentFirst"):
            if isinstance(self.syncCurrentFirst, bool) and self.syncCurrentFirst is True:
                logging.info("Synchronizing all configuration to Jasmin")
                self.syncCurrentFirst = False
            else:
                logging.info("Recieved notice of change in configurations")

            logging.info("from MongoDB cluster. Applies to:")
        if payload["change_type"] == "sync" and "sync_data" in payload:
            """ Sync """
            jasmin_proxy.syncAll(
                collection_data=payload["sync_data"]
            )

        if payload["change_type"] == "insert":
            """ Add """
            jasmin_proxy.add(
                module=payload["module"],
                sub_id=payload["sub_id"],
                options=payload["sub_data"]
            )

        if payload["change_type"] in ["update", "replace"]:
            """ Edit """
            jasmin_proxy.edit(
                module=payload["module"],
                sub_id=payload["sub_id"],
                options=payload["sub_data"]
            )

        if payload["change_type"] == "delete":
            """ Remove """
            jasmin_proxy.remove(
                module=payload["module"],
                sub_id=payload["sub_id"]
            )

        logging.info("Finished applying configurations")



import os
from .mongodb import MongoDB
import logging

NODEFAULT: str = "REQUIRED: NO_DEFAULT"
DEFAULT_SYNC_CURRENT_FIRST: bool = True
DEFAULT_JASMIN_CLI_HOST: str = '127.0.0.1'
DEFAULT_JASMIN_CLI_PORT: int = 8990
DEFAULT_JASMIN_CLI_TIMEOUT: int = 30
DEFAULT_JASMIN_CLI_AUTH: bool = True
DEFAULT_JASMIN_CLI_USERNAME: str = "jcliadmin"
DEFAULT_JASMIN_CLI_PASSWORD: str = "jclipwd"
DEFAULT_JASMIN_CLI_STANDARD_PROMPT: str = "jcli : "
DEFAULT_JASMIN_CLI_INTERACTIVE_PROMPT: str = "> "
DEFAULT_LOG_PATH: str = "/var/log/jasmin/"
DEFAULT_LOG_LEVEL: str = "INFO"


class ConfigurationStreamer:
    def __init__(
        self,
        mongo_connection_string: str,
        configuration_database: str,
        sync_current_first: bool = DEFAULT_SYNC_CURRENT_FIRST,
        jasmin_cli_host: str = DEFAULT_JASMIN_CLI_HOST,
        jasmin_cli_port: int = DEFAULT_JASMIN_CLI_PORT,
        jasmin_cli_timeout: int = DEFAULT_JASMIN_CLI_TIMEOUT,
        jasmin_cli_auth: bool = DEFAULT_JASMIN_CLI_AUTH,
        jasmin_cli_username: str = DEFAULT_JASMIN_CLI_USERNAME,
        jasmin_cli_password: str = DEFAULT_JASMIN_CLI_PASSWORD,
        jasmin_cli_standard_prompt: str = DEFAULT_JASMIN_CLI_STANDARD_PROMPT,
        jasmin_cli_interactive_prompt: str = DEFAULT_JASMIN_CLI_INTERACTIVE_PROMPT,
        logPath: str = DEFAULT_LOG_PATH,
        logLevel: str = DEFAULT_LOG_LEVEL
    ):

        self.MONGO_CONNECTION_STRING = mongo_connection_string

        self.MONGO_CONFIGURATION_DATABASE = configuration_database

        self.SYNC_CURRENT_FIRST = sync_current_first

        self.telnet_config: dict = {
            "host": jasmin_cli_host,
            "port": jasmin_cli_port,
            "timeout": jasmin_cli_timeout,
            "auth": jasmin_cli_auth,
            "username": jasmin_cli_username,
            "password": jasmin_cli_password,
            "standard_prompt": jasmin_cli_standard_prompt,
            "interactive_prompt": jasmin_cli_interactive_prompt
        }

        logFormatter = logging.Formatter(
            "%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
        rootLogger = logging.getLogger()
        rootLogger.setLevel(logLevel)

        consoleHandler = logging.StreamHandler()
        consoleHandler.setLevel(logLevel)
        consoleHandler.setFormatter(logFormatter)
        rootLogger.addHandler(consoleHandler)

        if not os.path.exists(logPath):
            os.makedirs(logPath)

        fileHandler = logging.FileHandler(
            f"{logPath.rstrip('/')}/jasmin_mongo_configuration.log")
        fileHandler.setLevel(logLevel)
        fileHandler.setFormatter(logFormatter)
        rootLogger.addHandler(fileHandler)

    def start(self):

        logging.info("*********************************************")
        logging.info("::Jasmin MongoDB Configuration::")
        logging.info("")

        mongosource = MongoDB(connection_string=self.MONGO_CONNECTION_STRING,
                              database_name=self.MONGO_CONFIGURATION_DATABASE)
        if mongosource.startConnection() is True:
            mongosource.stream(
                telnet_config=self.telnet_config, syncCurrentFirst=self.SYNC_CURRENT_FIRST)


def startFromCLI():
    import argparse
    parser = argparse.ArgumentParser(
        description=f"Jasmin MongoDB Configuration, Links Jasmin SMS Gateway to MongoDB cluster's Change Stream (can be one node).")
    
    parser.add_argument('--connection-string', '-c', type=str, dest='mongo_connection_string',
        help='MongoDB Connection String, No default. ex. mongodb://mongoroot:mongopass@mongodb1:27017,mongodb2:27017,mongodb3:27017/?authSource=admin&replicaSet=rs',
        required=os.getenv("MONGO_CONNECTION_STRING") is None,
        default=os.getenv("MONGO_CONNECTION_STRING"))
    
    parser.add_argument('--configuration-database', '-db', type=str, dest='configuration_database',
        help='Configuration Database, No default. MongoDB database name where you have saved the jasmin configurations',
        required=os.getenv("MONGO_CONFIGURATION_DATABASE") is None,
        default=os.getenv("MONGO_CONFIGURATION_DATABASE"))
    
    parser.add_argument('--sync-current-first', '-sync', type=bool, dest='sync_current_first',
        help='Sync current configuration first, Enabled by default. if enabled, will sync the current configurations first before monitoring for any changes',
        choices=['yes', 'y', 'no', 'n'],
        required=False,
        default=bool(os.getenv("SYNC_CURRENT_FIRST", 'yes' if DEFAULT_SYNC_CURRENT_FIRST else 'no').lower() in ['yes', 'y']))
    
    parser.add_argument('--cli-host', '-H', type=str, dest='jasmin_cli_host',
        help='Jasmin CLI Host, Default: "127.0.0.1". The hostname of the jasmin server',
        required=False,
        default=os.getenv("JASMIN_CLI_HOST", DEFAULT_JASMIN_CLI_HOST))
    
    parser.add_argument('--cli-port', '-P', type=int, dest='jasmin_cli_port',
        help='Jasmin CLI Port, Default: "8990". The port of the jasmin server cli',
        required=False,
        default=int(os.getenv("JASMIN_CLI_PORT", DEFAULT_JASMIN_CLI_PORT)))
    
    parser.add_argument('--cli-timeout', '-t', type=int, dest='jasmin_cli_timeout',
        help='Jasmin CLI Timeout, Default: "30". The timeout for the CLI connection. Should be increased if your network is not stable.',
        required=False,
        default=int(os.getenv("JASMIN_CLI_TIMEOUT", DEFAULT_JASMIN_CLI_TIMEOUT)))
    
    parser.add_argument('--cli-auth', '-a', type=bool, dest='jasmin_cli_auth',
        help='Jasmin CLI Auth, Enabled by default. if enabled, will use authentication for the telnet connection.',
        choices=['yes', 'y', 'no', 'n'],
        required=False,
        default=bool(os.getenv("JASMIN_CLI_AUTH", 'yes' if DEFAULT_JASMIN_CLI_AUTH else 'no').lower() in ['yes', 'y']))
    
    parser.add_argument('--cli-username', '-u', type=str, dest='jasmin_cli_username',
        help='Jasmin CLI Username, Default: "jcliadmin". The jasmin telnet cli username',
        required=False,
        default=os.getenv("JASMIN_CLI_USERNAME", DEFAULT_JASMIN_CLI_USERNAME))
    
    parser.add_argument('--cli-password', '-p', type=str, dest='jasmin_cli_password',
        help='Jasmin CLI Password, Default: "jclipwd". The jasmin telnet cli password',
        required=False,
        default=os.getenv("JASMIN_CLI_PASSWORD", DEFAULT_JASMIN_CLI_PASSWORD))
    
    parser.add_argument('--cli-standard-prompt', type=str, dest='jasmin_cli_standard_prompt',
        help='Jasmin CLI Standard Prompt, Default: "jcli : ". There shouldn\'t be a need to change this.',
        required=False,
        default=os.getenv("JASMIN_CLI_STANDARD_PROMPT", DEFAULT_JASMIN_CLI_STANDARD_PROMPT))
    
    parser.add_argument('--cli-interactive-prompt', type=str, dest='jasmin_cli_interactive_prompt',
        help='Jasmin CLI Interactive Prompt, Default: "> ". There shouldn\'t be a need to change this.',
        required=False,
        default=os.getenv("JASMIN_CLI_INTERACTIVE_PROMPT", DEFAULT_JASMIN_CLI_INTERACTIVE_PROMPT))
    
    parser.add_argument('--log-path', type=str, dest='logPath',
        help='Log Path, Default: "/var/log/jasmin/"',
        required=False,
        default=os.getenv("JASMIN_MONGO_CONFIGURATION_LOG_PATH", DEFAULT_LOG_PATH))
    
    parser.add_argument('--log-level', type=str, dest='logLevel',
        help='Log Level, Default: "INFO"',
        required=False,
        default=os.getenv("JASMIN_MONGO_CONFIGURATION_LOG_LEVEL", DEFAULT_LOG_LEVEL))

    configurationstreamer = ConfigurationStreamer(**vars(parser.parse_args()))

    configurationstreamer.start()
